// Copyright  2007-2014, Johan Hedborg, Linkping University
//                        Andreas Robinson, Linkping University
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from this
// software without specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

//////////////////////////////////////////////////////////////////////////
//		L I N E A R  A L G E B R A  3
// Desc: functionality for math operations between
//		 Vector3 and Matrix3x3
// 

#pragma once

#include "Vector3.hpp"
#include "Matrix3x3.hpp"
#include "Matrix3x4.hpp"

namespace cvl{

	//		Matrix3x3 * Vector3
	template<typename T>
	inline cvl::Vector3<T> operator*( const Matrix3x3<T> &A, const cvl::Vector3<T> &v )
	{
		return Vector3<T> (
			A.a00*v.x + A.a01*v.y + A.a02*v.z,
			A.a10*v.x + A.a11*v.y + A.a12*v.z,
			A.a20*v.x + A.a21*v.y + A.a22*v.z );
	}

	//		Scalar * Matrix3x3
	template<typename T>
	inline cvl::Matrix3x3<T> operator*(const T &s, const Matrix3x3<T> &A)
	{
		return Matrix3x3<T> (
			A.a00*s, A.a01*s, A.a02*s,
			A.a10*s, A.a11*s, A.a12*s,
			A.a20*s, A.a21*s, A.a22*s );
	}


	// Cross product
	template<typename T>
	inline cvl::Vector3<T> cross(const cvl::Vector3<T> &v1, const cvl::Vector3<T> &v2)
	{
		return Vector3<T> ( v1.y*v2.z - v1.z*v2.y,
							v1.z*v2.x - v1.x*v2.z,
							v1.x*v2.y - v1.y*v2.x);
	}

	// The Euclidean norm 
	template<typename T>
	inline T norm(const Vector3<T> &v1)
	{
		return sqrt( v1.x*v1.x + v1.y*v1.y + v1.z*v1.z );
	}


	// Squared Euclidean norm
	template<typename T>
	inline T squaredNorm(const Vector3<T> &v1)
	{
		return v1.x*v1.x + v1.y*v1.y + v1.z*v1.z;
	}


	// Dot product 
	template<typename T>
	inline T dot(const Vector3<T> &v1, const Vector3<T> &v2)
	{
		return ( v1.x*v2.x + v1.y*v2.y + v1.z*v2.z );
	}


	//		Matrix3x3 * Matrix3x3
	template<typename T>
	inline Matrix3x3<T> operator*( const Matrix3x3<T> &A, const Matrix3x3<T> &B )
	{
		return Matrix3x3<T>( 
			A.a00*B.a00 + A.a01*B.a10 + A.a02*B.a20,
			A.a00*B.a01 + A.a01*B.a11 + A.a02*B.a21,
			A.a00*B.a02 + A.a01*B.a12 + A.a02*B.a22,

			A.a10*B.a00 + A.a11*B.a10 + A.a12*B.a20,
			A.a10*B.a01 + A.a11*B.a11 + A.a12*B.a21,
			A.a10*B.a02 + A.a11*B.a12 + A.a12*B.a22,

			A.a20*B.a00 + A.a21*B.a10 + A.a22*B.a20,
			A.a20*B.a01 + A.a21*B.a11 + A.a22*B.a21,
			A.a20*B.a02 + A.a21*B.a12 + A.a22*B.a22 );
	}


	//Multiply  Matrix3x3 * Matrix3x4  to get a 3x4 as a result
    template<typename T>
	inline Matrix3x4<T> operator*( const Matrix3x3<T> &A, const Matrix3x4<T> &B )
	{
		
	  return Matrix3x4<T>(
		  A.a00*B.a00 + A.a01*B.a10 + A.a02*B.a20, A.a00*B.a01 + A.a01*B.a11 + A.a02*B.a21, A.a00*B.a02 + A.a01*B.a12 + A.a02*B.a22, A.a00*B.a03 + A.a01*B.a13 + A.a02*B.a23,
		  A.a10*B.a00 + A.a11*B.a10 + A.a12*B.a20, A.a10*B.a01 + A.a11*B.a11 + A.a12*B.a21, A.a10*B.a02 + A.a11*B.a12 + A.a12*B.a22, A.a10*B.a03 + A.a11*B.a13 + A.a12*B.a23,
		  A.a20*B.a00 + A.a21*B.a10 + A.a22*B.a20, A.a20*B.a01 + A.a21*B.a11 + A.a22*B.a21, A.a20*B.a02 + A.a21*B.a12 + A.a22*B.a22, A.a20*B.a03 + A.a21*B.a13 + A.a22*B.a23 );
	}


}
