// Copyright  2007-2014, Johan Hedborg, Linkping University
//                        Andreas Robinson, Linkping University
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from this
// software without specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

//////////////////////////////////////////////////////////////////////////
//		L I N E A R  A L G E B R A  4
// Desc: functionality for math operations between
//		 Vector4 and Matrix4x4
// 

#pragma once

#include "Vector4.hpp"
#include "Matrix3x4.hpp"
#include "Matrix4x4.hpp"


namespace cvl{

	// Set the 4x4 identity matrix 
	template<typename T>
	inline void setIdentity( Matrix4x4<T> &A)
	{
		A.a00 = 1.0; A.a01 = 0.0; A.a02 = 0.0; A.a03 = 0.0;
		A.a10 = 0.0; A.a11 = 1.0; A.a12 = 0.0; A.a13 = 0.0;
		A.a20 = 0.0; A.a21 = 0.0; A.a22 = 1.0; A.a23 = 0.0;
		A.a30 = 0.0; A.a31 = 0.0; A.a32 = 0.0; A.a33 = 1.0;
	}


	//		Matrix4x4 * Vector4
	template<typename T>
	inline cvl::Vector4<T> operator*( const Matrix4x4<T> &A, const cvl::Vector4<T> &v )
	{
		return Vector4<T> (
			A.a00*v.x + A.a01*v.y + A.a02*v.z + A.a03*v.w,
			A.a10*v.x + A.a11*v.y + A.a12*v.z + A.a13*v.w,
			A.a20*v.x + A.a21*v.y + A.a22*v.z + A.a23*v.w ,
			A.a30*v.x + A.a31*v.y + A.a32*v.z + A.a33*v.w );
	}


	//		Vector4 * Matrix4x4
	template<typename T>
	inline cvl::Vector4<T> operator*(const cvl::Vector4<T> &v, const Matrix4x4<T> &A)
	{
		return Vector4<T> (
			A.a00*v.x + A.a10*v.y + A.a20*v.z + A.a30*v.w,
			A.a01*v.x + A.a11*v.y + A.a21*v.z + A.a31*v.w,		
			A.a02*v.x + A.a12*v.y + A.a22*v.z + A.a32*v.w,
			A.a03*v.x + A.a13*v.y + A.a23*v.z + A.a33*v.w );
	}


	//		Scalar * Matrix3x4 
	template<typename T>
	inline cvl::Matrix3x4<T> operator*( const T &s, const Matrix3x4<T> &A )
	{
		return Matrix3x4<T> (
			A.a00*s, A.a01*s, A.a02*s, A.a03*s,
			A.a10*s, A.a11*s, A.a12*s, A.a13*s,
			A.a20*s, A.a21*s, A.a22*s, A.a23*s );
	}


	//		Matrix3x4 * Vector4
	template<typename T>
	inline Vector3<T> operator*( const Matrix3x4<T> &A, const Vector4<T> &v )
	{
		return Vector3<T> (
			A.a00*v.x + A.a01*v.y + A.a02*v.z + A.a03*v.w,
			A.a10*v.x + A.a11*v.y + A.a12*v.z + A.a13*v.w,
			A.a20*v.x + A.a21*v.y + A.a22*v.z + A.a23*v.w );
	}


	//		Vector3 * Matrix3x4
	template<typename T>
	inline Vector4<T> operator*(const Vector3<T> &v, const Matrix3x4<T> &A)
	{
		return Vector4<T> (
			A.a00*v.x + A.a10*v.y + A.a20*v.z,
			A.a01*v.x + A.a11*v.y + A.a21*v.z,		
			A.a02*v.x + A.a12*v.y + A.a22*v.z,
			A.a03*v.x + A.a13*v.y + A.a23*v.z );
	}


	//		Matrix4x4 * Matrix4x4
	template<typename T>
	inline Matrix4x4<T> operator*( const Matrix4x4<T> &A, const Matrix4x4<T> &B )
	{
		return Matrix4x4<T>( 
			A.a00*B.a00 + A.a01*B.a10 + A.a02*B.a20 + A.a03*B.a30,
			A.a00*B.a01 + A.a01*B.a11 + A.a02*B.a21 + A.a03*B.a31,
			A.a00*B.a02 + A.a01*B.a12 + A.a02*B.a22 + A.a03*B.a32,
			A.a00*B.a03 + A.a01*B.a13 + A.a02*B.a23 + A.a03*B.a33,


			A.a10*B.a00 + A.a11*B.a10 + A.a12*B.a20 + A.a13*B.a30,
			A.a10*B.a01 + A.a11*B.a11 + A.a12*B.a21 + A.a13*B.a31,
			A.a10*B.a02 + A.a11*B.a12 + A.a12*B.a22 + A.a13*B.a32,
			A.a10*B.a03 + A.a11*B.a13 + A.a12*B.a23 + A.a13*B.a33,


			A.a20*B.a00 + A.a21*B.a10 + A.a22*B.a20 + A.a23*B.a30,
			A.a20*B.a01 + A.a21*B.a11 + A.a22*B.a21 + A.a23*B.a31,
			A.a20*B.a02 + A.a21*B.a12 + A.a22*B.a22 + A.a23*B.a32,
			A.a20*B.a03 + A.a21*B.a13 + A.a22*B.a23 + A.a23*B.a33,


			A.a30*B.a00 + A.a31*B.a10 + A.a32*B.a20 + A.a33*B.a30,
			A.a30*B.a01 + A.a31*B.a11 + A.a32*B.a21 + A.a33*B.a31,
			A.a30*B.a02 + A.a31*B.a12 + A.a32*B.a22 + A.a33*B.a32,
			A.a30*B.a03 + A.a31*B.a13 + A.a32*B.a23 + A.a33*B.a33 );
	}


	//	Matrix3x4 / Scalar
	template<typename T>
	inline cvl::Matrix3x4<T> operator/( const Matrix3x4<T> &A, const T &s )
	{
		return Matrix3x4<T> (
			A.a00/s, A.a01/s, A.a02/s, A.a03/s, 
			A.a10/s, A.a11/s, A.a12/s, A.a13/s, 
			A.a20/s, A.a21/s, A.a22/s, A.a23/s );
	}


	// Calculate the L2 norm of a 3x4 matrix
    template<typename T>
	inline T norm(const Matrix3x4<T> &A)
	{
     return sqrt(A.a00*A.a00 + A.a01*A.a01 + A.a02*A.a02 + A.a03*A.a03
			   + A.a10*A.a10 + A.a11*A.a11 + A.a12*A.a12 + A.a13*A.a13 
			   + A.a20*A.a20 + A.a21*A.a21 + A.a22*A.a22 + A.a23*A.a23);
	}

	// Dot product 
	template<typename T>
	inline T dot(const Vector4<T> &v1, const Vector4<T> &v2)
	{
		return ( v1.x*v2.x + v1.y*v2.y + v1.z*v2.z + v1.w*v2.w );
	}


	// this can be optimized further !
	// by eliminate the copy to mat thingi
	template<typename T>
	inline Matrix4x4<T> inverse( const Matrix4x4<T> &M )
	{
		T* mat = new T[16];
		T* dst = new T[16];

		//Copy all of the elements into the linear array
		//int k=0;
		//for(int i=0; i<4; i++)
		//	for(int j=0; j<4; j++)
		//		mat[k++] = matrix[i][j];

		for(int i=0; i<16; i++)
			mat[i] = (&M.a00)[i];


		T* tmp = new T[12];

		// temp array for pairs
		T* src = new T[16];

		// array of transpose source matrix
		T det;

		/*determinant*/
		/*transposematrix*/
		for(int i=0;i<4;i++) //>
		{
			src[i]=mat[i*4];
			src[i+4]=mat[i*4+1];
			src[i+8]=mat[i*4+2];
			src[i+12]=mat[i*4+3];
		}

		// calculate pairs for first 8 elements (cofactors)
		tmp[0]=src[10]*src[15];
		tmp[1]=src[11]*src[14];
		tmp[2]=src[9]*src[15];
		tmp[3]=src[11]*src[13];
		tmp[4]=src[9]*src[14];
		tmp[5]=src[10]*src[13];
		tmp[6]=src[8]*src[15];
		tmp[7]=src[11]*src[12];
		tmp[8]=src[8]*src[14];
		tmp[9]=src[10]*src[12];
		tmp[10]=src[8]*src[13];
		tmp[11]=src[9]*src[12];

		// calculate first 8 elements (cofactors) 
		dst[0]=tmp[0]*src[5]+tmp[3]*src[6]+tmp[4]*src[7];
		dst[0]-=tmp[1]*src[5]+tmp[2]*src[6]+tmp[5]*src[7];
		dst[1]=tmp[1]*src[4]+tmp[6]*src[6]+tmp[9]*src[7];
		dst[1]-=tmp[0]*src[4]+tmp[7]*src[6]+tmp[8]*src[7];
		dst[2]=tmp[2]*src[4]+tmp[7]*src[5]+tmp[10]*src[7];
		dst[2]-=tmp[3]*src[4]+tmp[6]*src[5]+tmp[11]*src[7];
		dst[3]=tmp[5]*src[4]+tmp[8]*src[5]+tmp[11]*src[6];
		dst[3]-=tmp[4]*src[4]+tmp[9]*src[5]+tmp[10]*src[6];
		dst[4]=tmp[1]*src[1]+tmp[2]*src[2]+tmp[5]*src[3];
		dst[4]-=tmp[0]*src[1]+tmp[3]*src[2]+tmp[4]*src[3];
		dst[5]=tmp[0]*src[0]+tmp[7]*src[2]+tmp[8]*src[3];
		dst[5]-=tmp[1]*src[0]+tmp[6]*src[2]+tmp[9]*src[3];
		dst[6]=tmp[3]*src[0]+tmp[6]*src[1]+tmp[11]*src[3];
		dst[6]-=tmp[2]*src[0]+tmp[7]*src[1]+tmp[10]*src[3];
		dst[7]=tmp[4]*src[0]+tmp[9]*src[1]+tmp[10]*src[2];
		dst[7]-=tmp[5]*src[0]+tmp[8]*src[1]+tmp[11]*src[2];

		// calculate pairs for second 8 elements(cofactors)
		tmp[0]=src[2]*src[7];
		tmp[1]=src[3]*src[6];
		tmp[2]=src[1]*src[7];
		tmp[3]=src[3]*src[5];
		tmp[4]=src[1]*src[6];
		tmp[5]=src[2]*src[5];
		tmp[6]=src[0]*src[7];
		tmp[7]=src[3]*src[4];
		tmp[8]=src[0]*src[6];
		tmp[9]=src[2]*src[4];
		tmp[10]=src[0]*src[5];
		tmp[11]=src[1]*src[4];

		// calculate second 8 elements (cofactors)
		dst[8]=tmp[0]*src[13]+tmp[3]*src[14]+tmp[4]*src[15];
		dst[8]-=tmp[1]*src[13]+tmp[2]*src[14]+tmp[5]*src[15];
		dst[9]=tmp[1]*src[12]+tmp[6]*src[14]+tmp[9]*src[15];
		dst[9]-=tmp[0]*src[12]+tmp[7]*src[14]+tmp[8]*src[15];
		dst[10]=tmp[2]*src[12]+tmp[7]*src[13]+tmp[10]*src[15];
		dst[10]-=tmp[3]*src[12]+tmp[6]*src[13]+tmp[11]*src[15];
		dst[11]=tmp[5]*src[12]+tmp[8]*src[13]+tmp[11]*src[14];
		dst[11]-=tmp[4]*src[12]+tmp[9]*src[13]+tmp[10]*src[14];
		dst[12]=tmp[2]*src[10]+tmp[5]*src[11]+tmp[1]*src[9];
		dst[12]-=tmp[4]*src[11]+tmp[0]*src[9]+tmp[3]*src[10];
		dst[13]=tmp[8]*src[11]+tmp[0]*src[8]+tmp[7]*src[10];
		dst[13]-=tmp[6]*src[10]+tmp[9]*src[11]+tmp[1]*src[8];
		dst[14]=tmp[6]*src[9]+tmp[11]*src[11]+tmp[3]*src[8];
		dst[14]-=tmp[10]*src[11]+tmp[2]*src[8]+tmp[7]*src[9];
		dst[15]=tmp[10]*src[10]+tmp[4]*src[8]+tmp[9]*src[9];
		dst[15]-=tmp[8]*src[9]+tmp[11]*src[10]+tmp[5]*src[8];

		// calculate determinant
		det=src[0]*dst[0]+src[1]*dst[1]+src[2]*dst[2]+src[3]*dst[3];

		// calculate matrix inverse
		det=1/det;

		for(int j=0;j<16;j++) //>
			dst[j]*=det;

		//Copy everything into the output
		Matrix4x4<T> ret;
		//k = 0;
		//for(int i=0; i<4; i++)
		//	for(int j=0; j<4; j++)
		//		ret.matrix[i][j] = dst[k++];

		for(int i=0; i<16; i++)
			(&ret.a00)[i] = dst[i];

		return ret;
	}



}
