// Copyright © 2007-2014, Johan Hedborg, Linköping University
//                        Andreas Robinson, Linköping University
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from this
// software without specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

#include "stdafx.h"
#include "Vector2.hpp"
#include "LinearAlgebra3.hpp"
#include "MathUtilities.h"

using namespace cvl;

inline double sign(double value)
{
	if (value == 0.0)
		return 0.0;
	else
		if (value > 0.0)
			return 1.0;
	return -1.0;
}

// Useful reference: http://homel.vsb.cz/~dom033/predmety/parisLA/05_orthogonalization.pdf

//////// Householder QR
inline void qr_hh(const cvl::Matrix3x3D& A, cvl::Matrix3x3D& R, cvl::Matrix3x3D& Q)
{	
	Matrix3x3D I(1,0,0, 0,1,0, 0,0,1); // Identity matrix

	Q = I;
	R = A;

	{
		Vector3D x = R.col(0);
		Vector3D v = -sign(x(0)) * x.length() * Vector3D(1,0,0) - x;
		double Lv = v.length();

		if (Lv > 0) {
			v = v / Lv;

			Matrix3x3D P;

			// P = I - 2*v*v.transpose();

			P.a00 = 1 - 2 * v(0) * v(0);
			P.a01 =    -2 * v(0) * v(1);
			P.a02 =    -2 * v(0) * v(2);

			P.a10 =    -2 * v(1) * v(0);
			P.a11 = 1 - 2 * v(1) * v(1);
			P.a12 =    -2 * v(1) * v(2);

			P.a20 =    -2 * v(2) * v(0);
			P.a21 =    -2 * v(2) * v(1);
			P.a22 = 1 - 2 * v(2) * v(2);

			R = P * R;
			Q = Q * P;
		}
	}

	{
		Vector2D x = Vector2D(R(1,1), R(2,1));
		Vector2D v = -sign(x(0)) * x.length() * Vector2D(1,0) - x;
		double Lv = v.length();

		if (Lv > 0) {
			v = v / Lv;
			Matrix3x3D P = I;

			// P.bottomRightCorner<2,2>() -= 2 * v * v.transpose()
			P.a11 -= 2 * v(0) * v(0);
			P.a12 -= 2 * v(0) * v(1);
			P.a21 -= 2 * v(1) * v(0);
			P.a22 -= 2 * v(1) * v(1);

			R = P * R;
			Q = Q * P;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////
//                     r q 
//
// Orthogonal-triangular decomposition, A = R*Q, where R is an upper
// triangular matrix and Q is a unitary matrix.
void rq(const Matrix3x3D& A, Matrix3x3D& R, Matrix3x3D& Q)
{
	//A = A.transpose();
	//A = A(end:-1:1,end:-1:1);
	Matrix3x3D A2 = Matrix3x3D (
		A(2,2), A(1,2), A(0,2), 
		A(2,1), A(1,1), A(0,1), 
		A(2,0), A(1,0), A(0,0));

	// [Q,R] = qr(A);
	Matrix3x3D R2;
	Matrix3x3D Q2;

	qr_hh(A2, R2, Q2);

	//R = R.transpose();
	//R = R(end:-1:1,end:-1:1);
	R = Matrix3x3D(
		R2(2,2), R2(1,2), R2(0,2), 
		R2(2,1), R2(1,1), R2(0,1), 
		R2(2,0), R2(1,0), R2(0,0));

	// Q = Q.transpose();
	// Q = Q(end:-1:1,end:-1:1);
	Q = Matrix3x3D(
		Q2(2,2), Q2(1,2), Q2(0,2), 
		Q2(2,1), Q2(1,1), Q2(0,1), 
		Q2(2,0), Q2(1,0), Q2(0,0));
}

//////////////////////////////////////////////////////////////////////////////////////
//				P 2 K R t 
//
// Computes camera matrix K, rotation R, and translation t from
// projection matrix P. Relation:
//
//         P ~ K[R' -R'*t]
//
//   In:   P: 3x4 projection matrix
//   Out:  K: camera matrix, R: rotation matrix, t: translation vector
void P2KRt(const cvl::Matrix3x4D& P, cvl::Matrix3x3D& K, cvl::Matrix3x3D& R, cvl::Vector3D& t)
{

	Matrix3x3D A(
		P.a00, P.a01, P.a02, 
		P.a10, P.a11, P.a12, 
		P.a20, P.a21, P.a22);

	Vector3D p4(P.a03, P.a13, P.a23);

	rq(A, K, R);

	t = K.inverse() * p4;

	double P_multFactor = 1.0 / K(2,2);
	K *= P_multFactor;

	// K should have positive sign
	// D = diag(sign(diag(K)));
	Matrix3x3D D( sign(K(0,0)), 0.0, 0.0, 0.0, sign(K(1,1)), 0.0, 0.0, 0.0, sign(K(2,2)) );
	K = K*D;
	R = D*R;
	t = D*t;

	// R should have det(R)=1 (no mirroring)
	double detR = R.determinant();
	t = detR * t;
	R = detR * R;

	// change to P = [R' -R'*t] representation
	R = R.transpose();
	t = -R*t;
}