/*

Copyright © 2013, Johan Hedborg, Linköping University
                  Andreas Robinson, Linköping University
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#pragma once

//////////////////////////////////////////////////////////////////////////
//		C H O L E S K Y   D E C O M P O S I T I O N
// 
template<typename real, int dim>
class CholeskyDecomposition
{
	real L[dim*dim];
	real D[dim];
	real y[dim];

public:
	void solve(const real *pA, const real *b, real *x)
	{	
		D[0] = pA[0];

		for(int j = 0 ; j < dim-1 ; j++) 
		{
			L[dim*j+j] = real(1.0);
			real d = real(1.0) / D[j];

			for(int i = j+1 ; i < dim ; i++) {
				L[dim*i+j] = d;
				real c = pA[dim*i+j];
				for(int ss = 1 ; ss < j+1 ; ss++) {
					c -= D[ss-1] * L[dim*i+ss-1] * L[dim*j+ss-1];
				}

				L[dim*i+j] *= c;
			}

			D[j+1] = pA[dim*(j+1)+j+1];
			for(int ss = 1 ; ss < j+2 ; ss++) {
				D[j+1] -= D[ss-1] * L[dim*(j+1)+ss-1] * L[dim*(j+1)+ss-1];
			}
		}
		L[dim*dim-1] = real(1.0);

		// --- solve first L*y=b, results in a backwards subtitution ---
		y[0] = b[0];
		for (int e = 1; e < dim ; e++) {
			y[e] = b[e];
			for (int i = 0; i < e ; i++) {
				y[e] -= y[i]	* L[dim*e+i];
			}
		}

		// --- then solve DL'=y ---
		x[dim-1] = y[dim-1]/D[dim-1];
		for (int e = (dim-2); e >= 0; e--) {
			x[e] = y[e] / D[e];
			for (int i = (dim-1); i > e ; i--) {
				x[e] -= x[i] * L[dim*i+e];
			}
		}
	}
};

//////////////////////////////////////////////////////////////////////////
//		P O W E L L   D O G   L E G   S O L V E R
// 

#define DOGL_A_EQUALS_B(a, b)			for (int j = 0; j < nParams ; j++) a[j] = b[j];
#define DOGL_A_EQUALS_B_PLUS_C(a,b,c)	for (int j = 0; j < nParams ; j++) a[j] = b[j] + c[j];

template<typename real, int nParams, int nResiduals>
class PowellDogLeg
{
public:

	// calculate the residual vector r and the residual f = 0.5*sum(r.^2)
	virtual void getResidual(const real* x, real* r, real* f) = 0;

	/// calculate the Jacobian
	virtual int getJacobian( const real* x, real* J ) = 0;

	PowellDogLeg(void)
	{
		// options 
		radius = 0.0;						// if unset (i.e 0.0) it defaults to 0.5 * (1.0 + this->norm(x0))
		maxeval = 1000;						// max number of getResiduals iterations 
		setOptions( 1e-7, 1e-8, 1e-7, 20 );
		maxNoIterations = 1000;
		iterHistogram = std::vector<int>(maxeval, 0);

	}

	~PowellDogLeg(void)
	{
	}

	// set options
	void setRadius( real _radius )
	{
		radius = _radius;
	}

	void setOptions( real _tolg, real _tolx, real _tolr, int _maxeval )
	{
		tolg = _tolg;
		tolx = _tolx;
		tolr = _tolr;
		maxeval = _maxeval;
	}

	void setMaxNoIterations( int _maxNoIterations )
	{
		maxNoIterations = _maxNoIterations;
	}

	// the heavy functions is done 
	// get_r_f(x, r, &f)	: k times
	// getJ(x, Jk)			: nIterations+1 times 
	// GetHGN(Jk,r,hGN)		: nIterations+1 times 

	int calc( const real* x0, real* xOpt )
	{
		stop = -1;
		k = 1;

		real Delta = (radius != 0.0) ? radius :  0.5 * (1.0 + this->norm(x0));
		real f;		// residual 

		DOGL_A_EQUALS_B(x,x0);

		real nh;		// the norm for the step
		real nhGN;		// the norm for the step Gauss-Newton

		getResidual(x, r, &f);
		stop = getJacobian(x, JTk);

		getGrad(JTk, r, g);		// return g = -(J'*r)

		real ngi = normInf(nParams, g);

		real nu = 2.0;			
		real nx = norm(x);
		bool fact = true;		// do factorization
		stop = 0;				
		int reduce = 0;  
		real nstep = 0;
		nIterations = 0;		// nIterations 
		real dLpre;

		while (stop == 0)
		{
			if( ngi == 0x7ff0000000000000 )		// isinf
				stop = -5;
			else if( ngi <= tolg )
				stop = 1;
			else if( Delta <= tolx*(tolx + nx) )
				stop = 2;
			else if( normInf(nResiduals, r) <= tolr )
				stop = 3;
			else if( k >= maxeval || nIterations >= maxNoIterations )
				stop = 4;
			else
			{
				if(fact)
				{
					// get the Gauss-Newton step
					GetHGN(JTk, r, hGN);

					nhGN = norm(hGN);
					fact = false;
				}

				if( nhGN > Delta )		// Include gradient
				{
					nh = Delta;
					real ng = norm(g);  

					// alpha = (ng / norm(J*g))^2;
					real alpha = ng*ng / squaredNorm(JTk, g);

					DOGL_A_EQUALS_B(gn, alpha*g);

					real ngn = alpha*ng;
					real bb,gb,c,beta;

					if( ngn >= Delta )
					{
						// h = (Delta/ngn)*gn;
						DOGL_A_EQUALS_B(h, (Delta/ngn)*gn);

						dLpre = Delta*(2*ngn - Delta)/(2*alpha);
					}
					else
					{	// Dog leg

						DOGL_A_EQUALS_B_PLUS_C(b, hGN, -gn);

						bb = dot<nParams>(b, b);
						gb = dot<nParams>(gn, b);

						c = (Delta + ngn) * (Delta - ngn);

						if(gb > 0)
						{
							beta = c / (gb + sqrt(gb*gb + c * bb));
						}
						else
						{
							beta = (sqrt(gb*gb + c * bb) - gb) / bb;
						}
						// h = gn + beta*b;
						DOGL_A_EQUALS_B_PLUS_C(h, gn, beta*b);

						dLpre = 0.5*alpha * (1.0-beta)*(1.0-beta) * ng*ng + beta*(2.0-beta)*f;
					}
				}
				else
				{
					// h = hGN
					DOGL_A_EQUALS_B(h, hGN);

					nh = nhGN;
					dLpre = f;
					if( nh <= tolx*(tolx + nx) )
						stop = 2;
				}
			}
			if(stop == 0)
			{
				// xnew = x + h;
				DOGL_A_EQUALS_B_PLUS_C(xnew, x, h);

				// h = xnew - x;
				DOGL_A_EQUALS_B_PLUS_C(h, xnew, -x);

				nstep = norm(h);

				// dL = f - 0.5*norm(r + J*h)^2;
				real dL = f - (real)0.5 * squaredNorm_r_Jh(r, JTk, h);
				real fn;

				getResidual(xnew, rn, &fn);
				k = k + 1;
				real rho;		// 
				int updx;		// x updated

				if(stop == 0)
				{
					real df = f - fn;
					if( df > 0 && dL > 0 )                        // Update x
					{
						nIterations = nIterations + 1;  
						updx = 1;
						DOGL_A_EQUALS_B(x, xnew);
						f = fn;
						getJacobian(x, JTk);

						for (int i = 0; i < nResiduals ; i++)
							r[i] = rn[i];

						fact = true;

						// g = -(J'*r);
						getGrad(JTk, r, g);
						ngi = normInf(nParams, g);
						rho = dL / df;
					}
					else
					{
						rho = -1;  
						updx = 0; 
					}

					// Update Delta
					if( fabs(rho-1.0) < 0.2 && nh > Delta/3.0 && reduce <= 0 )
					{
						Delta = 3.0 * Delta;
						nu = 2.0;
						reduce = 0;
					}
					else if( rho < 0.25 )
					{
						Delta = Delta / nu;
						nu = 2.0 * nu;
						reduce = 2;
					}
					else
						reduce = reduce - 1;


					// If Trace we store some iteration values 
					if(Trace && updx)
					{
						// X(:,kit+1) = xnew;   
						// perf(:,kit+1) = [fn ngi Delta]';
					}

					if( k > maxeval || nIterations >= maxNoIterations )
						stop = 3;

				}
			}
		}

		DOGL_A_EQUALS_B(xOpt, x);

		iterHistogram[k-1] += 1;

		return stop;
	}


	void printInfo(void)
	{
		printf("Iterations: %d\n", nIterations);
		printf("Cost function calls:  %d\n", k);

		//			stop = 1 :  Stopped by small gradient
		//                 2 :  Stopped by small x-step
		//                 3 :  Stopped by small r-vector
		//                 4 :  No. of iteration steps exceeded
		//                -1 :  x is not a real valued vector
		//                -2 :  r is not a real valued column vector
		//                -3 :  J is not a real valued matrix
		//                -4 :  Dimension mismatch in x, r, J
		//                -5 :  Overflow during computation 
		switch(stop)
		{
		case 1:
			printf("Stopped by small gradient\n"); break;
		case 2:
			printf("Stopped by small x-step\n"); break;
		case 3:
			printf("Stopped by small r-vector\n"); break;
		case 4:
			printf("No. of function evaluations exceeded\n"); break;
		case -1:
			printf("Error : x is not a real valued vector\n"); break;
		case -2:
			printf("Error : r is not a real valued column vector\n"); break;
		case -3:
			printf("Error : J is not a real valued matrix\n"); break;
		case -4 :
			printf("Error : Dimension mismatch in x, r, J\n"); break;
		case -5 :
			printf("Error : Overflow during computation\n"); break;
		default:
			printf("not recognized\n"); break;
		}
	}

	int stop;				// interrupt code 
	int k;					// 
	int nIterations;		// number of iterations from last run 

	std::vector<int> iterHistogram;	// Histogram: number of points per number of iterations. 

private:

	CholeskyDecomposition<real, 3> cholesky;

	static const bool Trace = false;

	// norm(J*g)^2
	real squaredNorm(real* JT, real* g)
	{
		real ret = 0;
		for (int i = 0; i < nResiduals; i++) {
			real a = 0;
			for (int j = 0; j < nParams; j++) {
				a += JT[nResiduals*j+i] * g[j];
			}
			ret += a*a;
		}

		return ret;
	}

	// return g = -(J'*r)
	void getGrad( const real* JT, const real* r, real* g )
	{
		for (int i = 0; i < nParams; i++) {
			g[i] = -dot<nResiduals>(&JT[nResiduals*i], r);
		}
	}

	// Get the Gauss-Newton step (hGN)
	// hGN = -(J'*J)^-1 * J'*r
	void GetHGN( const real* JT, const real* r, real* hGN )
	{
		MultAAT<nParams, nResiduals>(JT, pH);
		Mult<nParams, nResiduals>(JT, r, JTr);
		cholesky.solve(pH, JTr, hGN);

		DOGL_A_EQUALS_B(hGN, -hGN);
	}

	// returns norm(r + J*h)^2;
	real squaredNorm_r_Jh( const real* r, const real* JT, const real* h )
	{
		real ret = 0;
		for (int i = 0; i < nResiduals; i++) {
			real a = r[i];
			for (int j = 0; j < nParams; j++) {
				a += JT[nResiduals*j+i] * h[j];
			}

			ret += a*a;		
		}
		return ret;
	}

	// Various generic linear algebra functions

	// L2 norm
	real norm(const real* v)
	{
		return sqrt(dot<nParams>(v, v));
	}

	// L-infinity norm
	real normInf(int n, real* v)
	{
		real maxValue = fabs(v[0]);

		for (int i = 1; i < n ; i++) {
			if(fabs(v[i]) > maxValue) {
				maxValue = fabs(v[i]);
			}
		}

		return maxValue;
	}

	// s = a*b, where a and b are Nx1
	template<int N>
	inline real dot(const real* a, const real* b)
	{
		real s = 0.0;
		for (int i = 0; i < N; i++) {
			s += a[i] * b[i];
		}
		return s;
	}

	// b=A*x, where A is a MxN matrix, x is Nx1 and b is Mx1
	template<int M, int N>
	inline void Mult(const real* A, const real* x, real* b)
	{
		for (int i = 0; i < M ; i++) {
			b[i] = dot<N>(&A[i*N], x);
		}
	}

	// B = A*A', where A is MxN and B is MxM
	template<int M, int N>
	inline void MultAAT(const real* A, real* B)
	{

		for (int i = 0; i < M ; i++) {
			for (int j = i; j < M ; j++) {
				B[M*i+j] = dot<N>(&A[N*i], &A[N*j]);
				B[M*j+i] = B[M*i+j];
			}
		}
	};

	// Solver options

	real radius;
	real tolg;
	real tolx;
	real tolr;
	int maxeval;					// max number of getResiduals iterations 
	int maxNoIterations;			// max nIterations : max number of Jacobian and LinearSolver calls

	// Work variables

	real pH[nParams*nParams];		// temporary area when solving the normal equations

	real r[nResiduals];				// the residual vector
	real rn[nResiduals];			//

	real Jk[nResiduals*nParams];	// current Jacobian
	real JTk[nParams*nResiduals];	// Jacobian transposed
	real JTr[nParams];				// JT*r

	real g[nParams];				// gradient vector
	real gn[nParams];

	real x[nParams];				// parameter vector
	real xnew[nParams];				//

	real h[nParams];				// step
	real hGN[nParams];				// step Gauss-Newton

	real b[nParams];
};

#undef DOGL_A_EQUALS_B
#undef DOGL_A_EQUALS_B_PLUS_C

//////////////////////////////////////////////////////////////////////////
//		P O W E L L   D O G   L E G   T R I A N G U L A T O R
// 

template<typename real, int nParams = 3, int nResiduals = 6, int nImgPoints = 3> // Only change the 'real' template parameter.
class DogLegTriangulator : public PowellDogLeg<double, nParams, nResiduals>
{
	double ut[nResiduals];
	double un_vn[nResiduals];
	double vt[nResiduals];
	double vn[nResiduals];

	const double *m_Ps;
	const double *m_imgPts;

	// Get the residual vector (r), and the residual function value (f)
	void getResidual( const real* x, real* r, real* f )
	{
		(*f) = 0;

		for (int i = 0; i < nImgPoints; i++)
		{
			const real *P = m_Ps + i * 12;
			const real *img = m_imgPts + i * 2;

			ut[i] = P[4*0+0]*x[0] + P[4*0+1]*x[1] + P[4*0+2]*x[2] + P[4*0+3]; // P(1,:)*[x ;  1];
			vt[i] = P[4*1+0]*x[0] + P[4*1+1]*x[1] + P[4*1+2]*x[2] + P[4*1+3]; // P(2,:)*[x ;  1];
			// vn same as un
			un_vn[i] = P[4*2+0]*x[0] + P[4*2+1]*x[1] + P[4*2+2]*x[2] + P[4*2+3]; // P(3,:)*[x ;  1];

			real ru = ut[i] / un_vn[i] - img[0];
			real rv = vt[i] / un_vn[i] - img[1];

			r[2*i+0] = ru;
			r[2*i+1] = rv;

			(*f) += ru*ru + rv*rv;
		}

		(*f) *= 0.5;
	}

	// Get the Jacobian (J)
	int getJacobian(const real* x, real* J)
	{
		for (int i = 0; i < nImgPoints ; i++)
		{
			const real *P = m_Ps + i * 12;
			const real *img = m_imgPts + i * 2;

			real div = 1.0 / (un_vn[i]*un_vn[i]);

			real dudx = (un_vn[i]*P[4*0+0] - ut[i]*P[4*2+0]) * div; 
			real dudy = (un_vn[i]*P[4*0+1] - ut[i]*P[4*2+1]) * div; 
			real dudz = (un_vn[i]*P[4*0+2] - ut[i]*P[4*2+2]) * div; 

			real dvdx = (un_vn[i]*P[4*1+0] - vt[i]*P[4*2+0]) * div;
			real dvdy = (un_vn[i]*P[4*1+1] - vt[i]*P[4*2+1]) * div;
			real dvdz = (un_vn[i]*P[4*1+2] - vt[i]*P[4*2+2]) * div;

			J[nResiduals*0+2*i+0] = dudx;
			J[nResiduals*0+2*i+1] = dvdx;

			J[nResiduals*1+2*i+0] = dudy;
			J[nResiduals*1+2*i+1] = dvdy;

			J[nResiduals*2+2*i+0] = dudz;
			J[nResiduals*2+2*i+1] = dvdz;
		}

		return 0;
	}
public:
	// Ps: Array of three 3x4 projection matrices in row-major order
	// imgPts: Array of three 2-vectors
	void setInput(const real *Ps, const real* imgPts)
	{
		m_Ps = Ps;
		m_imgPts = imgPts;
	}
};
