// Copyright © 2007-2014, Johan Hedborg, Linköping University
//                        Andreas Robinson, Linköping University
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from this
// software without specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

#pragma once

#include <vector>

#include "Vector2.hpp"
#include "Vector3.hpp"
#include "Matrix3x3.hpp"
#include "Matrix3x4.hpp"

#include "LinearAlgebra3.hpp"
#include "LinearAlgebra4.hpp"

namespace cvl{

	// convert to Vector2 homogeneous coordinates (Vector3)
	template<typename T>
	inline Vector3<T> toHomogeneous( const Vector2<T> &point )
	{
		return Vector3<T>( point.x, point.y, (T)1.0 );
	}

	// convert to Vector3 homogeneous coordinates (Vector4)
	template<typename T>
	inline Vector4<T> toHomogeneous( const Vector3<T> &point )
	{
		return Vector4<T>( point.x, point.y, point.z, (T)1.0 );
	}

	// convert to Vector2 homogeneous coordinates (Vector3)
	template<typename T>
	inline Vector2<T> toInhomogeneous( const Vector3<T> &point )
	{
		return Vector2<T>( point.x/point.z, point.y/point.z );
	}

	// convert to Vector3 homogeneous coordinates (Vector4)
	template<typename T>
	inline Vector3<T> toInhomogeneous( const Vector4<T> &point )
	{
		return Vector3<T>( point.x/point.w, point.y/point.w, point.z/point.w );
	}

	// convert to Matrix4x4 homogeneous coordinates from a Matrix3x4 projecting matrix
	template<typename T>
	inline Matrix4x4<T> toHomogeneous( const Matrix3x4<T> &C )
	{
		return Matrix4x4<T>( C.a00, C.a01, C.a02, C.a03, C.a10, C.a11, C.a12, C.a13, C.a20, C.a21, C.a22, C.a23, 0.0, 0.0, 0.0, 1.0 );
	}

	// convert to Matrix4x4 homogeneous coordinates from a Matrix3x4 projecting matrix
	template<typename T>
	inline Matrix3x4<T> toInhomogeneous( const Matrix4x4<T> &C )
	{
		double s = C.a33;
		return Matrix3x4<T>( C.a00/s, C.a01/s, C.a02/s, C.a03/s, C.a10/s, C.a11/s, C.a12/s, C.a13/s, C.a20/s, C.a21/s, C.a22/s, C.a23/s );
	}
}

