// Copyright © 2007-2014, Johan Hedborg, Linköping University
//                        Andreas Robinson, Linköping University
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from this
// software without specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

#include "stdafx.h"
#include "Triangulator.h"
#include "ProjectiveAlgebra.hpp"
#include "Time.h"
#include <limits>


using namespace std;
using namespace cvl;

extern double g_SolverSetupTime;
extern double g_SolverRuntime;

/*
Calculate the line segment PaPb that is the shortest route between
two lines P1P2 and P3P4. Calculate also the values of mua and mub where
Pa = P1 + mua (P2 - P1)
Pb = P3 + mub (P4 - P3)
Return zero if no solution exists.
*/
int Triangulator::LineLineIntersect( Vector3D& p1, Vector3D& p2,Vector3D& p3,Vector3D& p4,Vector3D& pa,Vector3D& pb, double& mua, double& mub)
{
	Vector3D p13,p43,p21;
	double d1343,d4321,d1321,d4343,d2121;
	double numer,denom;

	const double eps = std::numeric_limits<double>::epsilon();

	p13.x = p1.x - p3.x;
	p13.y = p1.y - p3.y;
	p13.z = p1.z - p3.z;
	p43.x = p4.x - p3.x;
	p43.y = p4.y - p3.y;
	p43.z = p4.z - p3.z;
	if (fabs(p43.x) < eps && fabs(p43.y) < eps && fabs(p43.z) <eps)
		return 0;
	p21.x = p2.x - p1.x;
	p21.y = p2.y - p1.y;
	p21.z = p2.z - p1.z;
	if (fabs(p21.x) < eps && fabs(p21.y) < eps && fabs(p21.z) < eps)
		return 0;

	d1343 = p13.x * p43.x + p13.y * p43.y + p13.z * p43.z;
	d4321 = p43.x * p21.x + p43.y * p21.y + p43.z * p21.z;
	d1321 = p13.x * p21.x + p13.y * p21.y + p13.z * p21.z;
	d4343 = p43.x * p43.x + p43.y * p43.y + p43.z * p43.z;
	d2121 = p21.x * p21.x + p21.y * p21.y + p21.z * p21.z;

	denom = d2121 * d4343 - d4321 * d4321;
	if (fabs(denom) < eps)
		return 0;
	numer = d1343 * d4321 - d1321 * d4343;

	mua = numer / denom;
	mub = (d1343 + d4321 * (mua)) / d4343;

	pa.x = p1.x + mua * p21.x;
	pa.y = p1.y + mua * p21.y;
	pa.z = p1.z + mua * p21.z;
	pb.x = p3.x + mub * p43.x;
	pb.y = p3.y + mub * p43.y;
	pb.z = p3.z + mub * p43.z;

	return 1;
}


// 
Vector3D Triangulator::TriangulateMidpoint(const Matrix3x4D& C0, const Matrix3x4D& C1, const Vector2D& x1, const Vector2D& x2)
{
	Vector3D pa, pb;
	double mua, mub;

	cvl::Matrix3x3D K0, R0, K1, R1;
	Vector3D c0, c1;
	P2KRt(C0, K0, R0, c0);
	P2KRt(C1, K1, R1, c1);
	Matrix3x3D K0inv = K0.inverse();
	Matrix3x3D K1inv = K1.inverse();

	Matrix3x4D C0n = K0inv * C0;
	Matrix3x4D C1n = K1inv * C1;
	Vector2D x1n = toInhomogeneous(K0inv * toHomogeneous(x1));
	Vector2D x2n = toInhomogeneous(K1inv * toHomogeneous(x2));

	Vector3D x1c1;
	x1c1.x = C0n.a00 * (x1n.x-C0n.a03) + C0n.a10 * (x1n.y-C0n.a13) + C0n.a20 * (1.0-C0n.a23);
	x1c1.y = C0n.a01 * (x1n.x-C0n.a03) + C0n.a11 * (x1n.y-C0n.a13) + C0n.a21 * (1.0-C0n.a23);
	x1c1.z = C0n.a02 * (x1n.x-C0n.a03) + C0n.a12 * (x1n.y-C0n.a13) + C0n.a22 * (1.0-C0n.a23);

	Vector3D x2c2;
	x2c2.x = C1n.a00 * (x2n.x-C1n.a03) + C1n.a10 * (x2n.y-C1n.a13) + C1n.a20 * (1.0-C1n.a23);
	x2c2.y = C1n.a01 * (x2n.x-C1n.a03) + C1n.a11 * (x2n.y-C1n.a13) + C1n.a21 * (1.0-C1n.a23);
	x2c2.z = C1n.a02 * (x2n.x-C1n.a03) + C1n.a12 * (x2n.y-C1n.a13) + C1n.a22 * (1.0-C1n.a23);

	LineLineIntersect(c0, x1c1, c1, x2c2, pa, pb, mua, mub);

	return pa + 0.5*(pb-pa);
}


// this can be done faster, FIXME when needed!!
int Triangulator::TriangulateMidpoint( const double* C0, const double* C1, const double* img0Point2D, const cvl::Vector2D& img1Point2D , double* resultPoint3D )
{
	Matrix3x4D _C0(C0);
	Matrix3x4D _C1(C1);
	Vector2D _x0(img0Point2D);
	Vector2D _x1(img1Point2D);

	Vector3D res = TriangulateMidpoint(_C0,_C1,_x0,_x1);

	resultPoint3D[0] = res[0];
	resultPoint3D[1] = res[1];
	resultPoint3D[2] = res[2];

	return 1;
}


cvl::Vector3D Triangulator::calc( const std::vector<cvl::Matrix3x4D>& Ps, const std::vector<cvl::Vector2D>& imgPoint )
{
	assert(Ps.size() == 3);
	assert(imgPoint.size() == 3);

	int64_t t0 = getTickCount();

	cvl::Vector3D point3D;

	m_imgPoint = imgPoint;
	m_Ps = Ps;

	solver.setOptions( 1e-5, 1e-6, 1e-5, 100 );
	solver.setInput(&m_Ps[0][0], &m_imgPoint[0][0]);
	
	int64_t t1 = getTickCount();

	// Initialize with a mid point triangulation from the first and last camera
	TriangulateMidpoint(&m_Ps[0][0], &(m_Ps.back())[0], &m_imgPoint[0][0], &(m_imgPoint.back())[0], &xInit[0]);
	solver.calc(&xInit[0], &point3D[0]);

	int64_t t2 = getTickCount();
	g_SolverSetupTime = (t1 - t0) / getTickFrequency();
	g_SolverRuntime = (t2 - t1) / getTickFrequency();

	return point3D;
}
