// Copyright  2007-2014, Johan Hedborg, Linkping University
//                        Andreas Robinson, Linkping University
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from this
// software without specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

//////////////////////////////////////////////////////////////////////////
//		V E C T O R  2
//
// Desc: A template vector class for 2 dimensional vectors
//

#pragma once

#include <math.h>
#include <string>
#include <stdio.h>

namespace cvl{

	//////////////////////////////////////////////////////////////////////////
	//		V E C T O R 2
	//
	template<typename T>
	class Vector2
	{
	public:

		// constructors
		Vector2<T>() {};
		Vector2<T>(T _x,T _y ) { x=_x; y=_y; }
		Vector2<T>(const Vector2<T>& v) {x = v.x; y = v.y;}
		Vector2<T>(const T* pV)	{ x = pV[0]; y = pV[1];};

		template<typename T2>	
		Vector2<T>(const Vector2<T2>& v) { x = (T)v.x; y = (T)v.y;}


		// member variable access
		T &operator()( const int& i )
			{ return this->v[i]; }

		T &operator[]( const int& i )
		{ return this->v[i]; }

		const T &operator()( const int& i ) const 
			{ return this->v[i]; }

		const T &operator[]( const int& i ) const 
			{ return v[i]; }


		// get the pointer to the first element 
		T* getPtr(void) { return &v[0]; } 
		const T* getConstPtr(void) const { return &v[0]; } 

		// size
		// unsigned int size(void) { return &v[0]; } 
		const unsigned int size(void) const { return 2; } 

		// arithmetic operations affecting the class instance
		void operator+=(const Vector2<T>& v)
			{ x+=v.x; y+=v.y; }

		void operator-=(const Vector2<T>& v)
			{ x-=v.x; y-=v.y; }

		void operator*=(const Vector2<T>& v)
			{ x*=v.x; y*=v.y; }

		void operator/=(const Vector2<T>& v)
			{ x/=v.x; y/=v.y; }

		void operator*=(const T& scalar)
			{ x*=scalar; y*=scalar;  }

		void operator/=(const T& scalar)
			{ x/=scalar; y/=scalar; }


		// length() : Returns the length of the vector (norm 2)
		// is norm2 a better name!?
		inline T length(void) const;

		// squaredLength() : Returns the length of the vector
		inline T squaredLength(void) const;

		// normalized() : Returns the current vector, normalized
		inline Vector2<T> normalized(void) const;

		// normalize() : Normalize current vector
		inline void normalize(void);

		// arithmetic operations
		inline Vector2<T> operator+( const Vector2<T>& v ) const;
		inline Vector2<T> operator-( const Vector2<T>& v ) const;
		inline Vector2<T> operator*( const Vector2<T>& v ) const;
		inline Vector2<T> operator/( const Vector2<T>& v ) const;
		inline Vector2<T> operator*( const T& s ) const;
		inline Vector2<T> operator/( const T& s ) const;
		inline Vector2<T> operator-() const;
		inline const Vector2<T>& operator=(const Vector2<T>& v);

		// operator == 
		inline bool operator==( const Vector2<T>& a ) const;

		// toString() : replaced whit operator <<
		// std::string toString(void) const;

		// member variables
		union {
			struct {
				T x,y;
			};		
			T v[2];
		};

	};

	//////////////////////////////////////////////////////////////////////////
	// member functions

	// length()
	template<typename T>
	inline T Vector2<T>::length(void) const
	{
		return sqrt(x*x+y*y);
	}

	// squaredLength()
	template<typename T>
	inline T Vector2<T>::squaredLength(void) const
	{
		return x*x+y*y;
	}

	// normalized()
	template<typename T>
	inline Vector2<T> Vector2<T>::normalized(void) const
	{
		T l=length();
		return Vector2<T>(x/l,y/l);
	}

	// normalize()
	template<typename T>
	inline void Vector2<T>::normalize(void)
	{
		T l=length();
		x/=l;
		y/=l;
	}

	// operator +
	template<typename T>
	inline Vector2<T> Vector2<T>::operator+( const Vector2<T>& v ) const
	{
		return  Vector2<T>( x+v.x, y+v.y );
	}

	// operator -
	template<typename T>
	inline Vector2<T> Vector2<T>::operator-( const Vector2<T>& v ) const
	{
		return  Vector2<T>( x-v.x, y-v.y );
	}

	// operator *
	template<typename T>
	inline Vector2<T> Vector2<T>::operator*( const Vector2<T>& v ) const
	{
		return  Vector2<T>( x*v.x, y*v.y );
	}

	// operator /
	template<typename T>
	inline Vector2<T> Vector2<T>::operator/( const Vector2<T>& v ) const
	{
		return  Vector2<T>( x/v.x, y/v.y );
	}

	// operator * for scalars
	template<typename T>
	inline Vector2<T> Vector2<T>::operator*( const T& s ) const
	{
		return  Vector2<T>( x*s, y*s );
	}

	// operator / for scalars
	template<typename T>
	inline Vector2<T> Vector2<T>::operator/( const T& s ) const
	{
		return  Vector2<T>( x/s, y/s );
	}

	// operator - ( -v )
	template<typename T>
	inline Vector2<T> Vector2<T>::operator-() const
	{
		return  Vector2<T>( -x, -y );
	}

	// assignment operator
	template<typename T>
	inline const Vector2<T>& Vector2<T>::operator=(const Vector2<T>& v)
	{
		x=v.x;  y=v.y;
		return *this;
	}

	// operator ==
	template<typename T>
	inline bool Vector2<T>::operator==( const Vector2<T>& v ) const
	{
		return  (v.x==x)&&(v.y==y);
	}

	// toString() PS!! is replaced whit operator<<
	//template<typename T>
	//std::string Vector2<T>::toString(void) const
	//{
	//	char str[256];
	//	sprintf(str, "(%g,%g)",x,y);
	//	return std::string(str);
	//}


	//////////////////////////////////////////////////////////////////////////
	// functions outside the class

	//	Scalar * Vector2
	template<typename T>
	inline cvl::Vector2<T> operator*( const T& scalar, const cvl::Vector2<T> &v )
	{
		return Vector2<T> ( scalar*v.x, scalar*v.y );
	}

	//	Scalar / Vector2
	template<typename T>
	inline cvl::Vector2<T> operator/( const T& scalar, const cvl::Vector2<T> &v )
	{
		return Vector2<T> ( scalar/v.x, scalar/v.y );
	}

	// operator <<
	template<typename T>
	std::ostream& operator<<(std::ostream& out, const Vector2<T>& v)
	{
		// the std::string is for GCC compatibility
		out << std::string("[") << v.x << std::string(", ") << v.y << std::string("]");
		return out;
	}
	
	//////////////////////////////////////////////////////////////////////////
	// Type definition and template instantiation

#ifdef _WIN32
	// instantiation template for some types 
	Vector2<float>;
	Vector2<double>;
#endif

	// typedef Vector2 Point2;	// does not work, do we need someting like this?

	typedef Vector2<float> Point2F;
	typedef Vector2<double> Point2D;

	typedef Vector2<float> Vector2F;
	typedef Vector2<double> Vector2D;
}
