// Copyright  2007-2014, Johan Hedborg, Linkping University
//                        Andreas Robinson, Linkping University
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from this
// software without specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

//////////////////////////////////////////////////////////////////////////
//		V E C T O R  3
//
// Desc: A template vector class for 3 dimensional vectors
//

#pragma once

#include <math.h>
#include <string>
#include <stdio.h>
#include <iostream>
//#include "Matrix3x4.hpp"
//#include "Matrix3x3.hpp"



namespace cvl{

	//////////////////////////////////////////////////////////////////////////
	//		V E C T O R 3
	//
	template<typename T>
	class Vector3
	{
	public:

		
		// constructors
		Vector3<T>() {}
		// Vector3<T>(T _value){x=_value; y=_value; z=_value; }
		Vector3<T>(T _x,T _y, T _z) { x=_x; y=_y; z=_z; }
		Vector3<T>(const Vector3<T>& v)	{ x = v.x; y = v.y; z = v.z;};
		Vector3<T>(const T* pV)	{ x = pV[0]; y = pV[1]; z = pV[2];};

		template<typename T2>	
		Vector3<T>(const Vector3<T2>& v) { x = (T)v.x; y = (T)v.y; z = (T)v.z;};


		// member variable access
		T &operator()( const int& i )
			{ return this->v[i]; }

		T &operator[]( const int& i )
			{ return this->v[i]; }

		const T &operator()( const int& i ) const 
			{ return this->v[i]; }

		const T &operator[]( const int& i ) const 
			{ return v[i]; }


		// get the pointer to the first elemet 
		T* getPtr(void) { return &v[0]; } 
		const T* getConstPtr(void) const { return &v[0]; } 
		
		// size
		// unsigned int size(void) { return &v[0]; } 
		const unsigned int size(void) const { return 3; } 
		
		// arithmetic operations affecting the class instance
		void operator+=(const Vector3<T>& v)
			{ x+=v.x; y+=v.y; z+=v.z; }

		void operator-=(const Vector3<T>& v)
			{ x-=v.x; y-=v.y; z-=v.z; }

		// is this needed , there is a vector - scalar, think of this 
		//void operator-=(const T& v)
		//	{ x-=v; y-=v; z-=v; }

		void operator*=(const Vector3<T>& v)
			{ x*=v.x; y*=v.y; z*=v.z; }

		//void operator*=(const Matrix3x3<T>& Tr)
		//	{ x = x*Tr.a00 + y*Tr.a01 + z*Tr.a02; y = x*Tr.a10 + y*Tr.a11 + z*Tr.a12; z = x*Tr.a20 + y*Tr.a21 + z*Tr.a22; }

		//void operator*=(const Matrix3x4<T>& Tr)
		//	{ x = x*Tr.a00 + y*Tr.a01 + z*Tr.a02 + Tr.a03; y = x*Tr.a10 + y*Tr.a11 + z*Tr.a12 + Tr.a13; z = x*Tr.a20 + y*Tr.a21 + z*Tr.a22 + Tr.a23; }



		void operator/=(const Vector3<T>& v)
			{ x/=v.x; y/=v.y; z/=v.z; }

		void operator*=(const T& scalar)
			{ x*=scalar; y*=scalar; z*=scalar; }

		void operator/=(const T& scalar)
			{ x/=scalar; y/=scalar; z/=scalar; }


		// length() : returns the length of the vector
		inline T length(void) const;

		// squaredLength() : returns the squared length of the vector
		inline T squaredLength(void) const;

		// normalized() : Returns the current vector, normalized
		inline Vector3<T> normalized(void) const;

		// normalize() : Normalize current vector
		inline void normalize(void);

		// + - * / and == operators
		inline Vector3<T> operator+( const Vector3<T>& v ) const;
		inline Vector3<T> operator-( const Vector3<T>& v ) const;
		inline Vector3<T> operator-( const T& s ) const;
		inline Vector3<T> operator*( const Vector3<T>& v ) const;
		inline Vector3<T> operator/( const Vector3<T>& v ) const;
		inline Vector3<T> operator*( const T& s ) const;
		inline Vector3<T> operator/( const T& s ) const;
		inline Vector3<T> operator-() const;
		inline const Vector3<T>& operator=(const Vector3<T>& v);
		
		// assignment operator
		inline bool operator==( const Vector3<T>& a ) const;


		// toString() is replaced whit operator <<
		//std::string toString(void);

		// member variables
		union {
			struct {
				T x,y,z;
			};		
			T v[3];
		};


	};

	//////////////////////////////////////////////////////////////////////////
	// member functions

	// length()
	template<typename T>
	inline T Vector3<T>::length(void) const
	{
		return sqrt(x*x+y*y+z*z);
	}

	// squaredLength()
	template<typename T>
	inline T Vector3<T>::squaredLength(void) const
	{
		return (x*x+y*y+z*z);
	}

	// normalized()
	template<typename T>
	inline Vector3<T> Vector3<T>::normalized(void) const
	{
		T l=length();
		return Vector3<T>(x/l,y/l,z/l);
	}

	// normalize() 
	template<typename T>
	inline void Vector3<T>::normalize(void)
	{
		T l=length();
		x/=l;
		y/=l;
		z/=l;
	}

	// operator +
	template<typename T>
	inline Vector3<T> Vector3<T>::operator+( const Vector3<T>& v ) const
	{
		return  Vector3<T>( x+v.x, y+v.y, z+v.z );
	}

	// operator -
	template<typename T>
	inline Vector3<T> Vector3<T>::operator-( const Vector3<T>& v ) const
	{
		return  Vector3<T>( x-v.x, y-v.y, z-v.z );
	}
	
	template<typename T>
	inline Vector3<T> Vector3<T>::operator-( const T& s ) const
	{
		return  Vector3<T>( x-s, y-s, z-s );
	}

	// operator *
	template<typename T>
	inline Vector3<T> Vector3<T>::operator*( const Vector3<T>& v ) const
	{
		return  Vector3<T>( x*v.x, y*v.y, z*v.z );
	}

	// operator /
	template<typename T>
	inline Vector3<T> Vector3<T>::operator/( const Vector3<T>& v ) const
	{
		return  Vector3<T>( x/v.x, y/v.y, z/v.z );
	}

	// operator * for scalars
	template<typename T>
	inline Vector3<T> Vector3<T>::operator*( const T& s ) const
	{
		return  Vector3<T>( x*s, y*s, z*s );
	}

	// operator / for scalars
	template<typename T>
	inline Vector3<T> Vector3<T>::operator/( const T& s ) const
	{
		return  Vector3<T>( x/s, y/s, z/s );
	}

	// operator - ( -v )
	template<typename T>
	inline Vector3<T> Vector3<T>::operator-() const
	{
		return Vector3<T>(-x, -y, -z);
	}
	
	// assignment operator
	template<typename T>
	inline const Vector3<T>& Vector3<T>::operator=(const Vector3<T>& v)
	{
		x=v.x;  y=v.y;  z=v.z;
		return *this;
	}

	// operator ==
	template<typename T>
	inline bool Vector3<T>::operator==( const Vector3<T>& v ) const
	{
		return  (v.x==x)&&(v.y==y)&&(v.z==z);
	}

	// toString() is replaced whit operator <<
	//template<typename T>
	//std::string Vector3<T>::toString(void)
	//{
	//	char str[256];
	//	sprintf(str, "(%g,%g,%g)'",x,y,z);
	//	return std::string(str);
	//}


	//////////////////////////////////////////////////////////////////////////
	// functions outside the class

	//	Scalar * Vector3
	template<typename T>
	inline cvl::Vector3<T> operator*( const T& scalar, const cvl::Vector3<T> &v )
	{
		return Vector3<T> ( scalar*v.x, scalar*v.y, scalar*v.z );
	}
	
	// Scalar / Vector3
	template<typename T>
	inline cvl::Vector3<T> operator/( const T& scalar, const cvl::Vector3<T> &v )
	{
		return Vector3<T> ( scalar/v.x, scalar/v.y, scalar/v.z );
	}

	// operator << 
	// use print() from 


	//////////////////////////////////////////////////////////////////////////
	// Type definition and template instantiation

#ifdef _WIN32
	// instantiation for some types 
	Vector3<float>;
	Vector3<double>; 
#endif

	// typedef Vector3 Point3;	// does not work, do we need someting like this?
	typedef Vector3<float> Point3F;
	typedef Vector3<double> Point3D;

	typedef Vector3<float> Vector3F;
	typedef Vector3<double> Vector3D;

}
