/*

Copyright  2007-2013, Johan Hedborg, Linkping University
                       Andreas Robinson, Linkping University
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#pragma once

#include <math.h>
#include <string>
#include <stdio.h>

namespace cvl{

	//////////////////////////////////////////////////////////////////////////
	//		V E C T O R 4
	//
	template<typename T>
	class Vector4
	{
	public:

		// constructors
		Vector4<T>() {};
		Vector4<T>(T _x,T _y, T _z, T _w){ x=_x; y=_y; z=_z; w=_w; }
		Vector4<T>(const Vector4<T>& v) {x = v.x; y = v.y; z = v.z; w = v.w;}
		Vector4<T>(const T* pV)	{ x = pV[0]; y = pV[1]; z = pV[2]; w = pV[3];};

		template<typename T2>	
		Vector4<T>(const Vector4<T2>& v) { x = (T)v.x; y = (T)v.y; z = (T)v.z; w = (T)v.w;}


		// member variable access
		T &operator()( const int& i )
			{ return this->v[i]; }

		T &operator[]( const int& i )
			{ return this->v[i]; }

		const T &operator()( const int& i ) const 
		{ return this->v[i]; }

		const T &operator[]( const int& i ) const 
		{ return v[i]; }


		// get the pointer to the first elemet
		T* getPtr(void) { return &v[0]; } 
		const T* getConstPtr(void) const { return &v[0]; } 


		// size
		// unsigned int size(void) { return &v[0]; } 
		const unsigned int size(void) const { return 4; } 


		// arithmetic operations affecting the class instance
		void operator+=(const Vector4<T>& v)
			{ x+=v.x; y+=v.y; z+=v.z;  w+=v.w; }

		void operator-=(const Vector4<T>& v)
			{ x-=v.x; y-=v.y; z-=v.z; w-=v.w; }

		void operator*=(const Vector4<T>& v)
			{ x*=v.x; y*=v.y; z*=v.z; w*=v.w; }

		void operator/=(const Vector4<T>& v)
			{ x/=v.x; y/=v.y; z/=v.z; w/=v.w; }

		void operator*=(const T& scalar)
			{ x*=scalar; y*=scalar; z*=scalar; w*=scalar; }

		void operator/=(const T& scalar)
			{ x/=scalar; y/=scalar; z/=scalar; w/=scalar; }


		// length() : Returns the length of the vector
		inline T length(void) const;

		// squaredLength() : Returns the length of the vector
		inline T squaredLength(void) const;

		// normalized() : Returns the current vector, normalized
		inline Vector4<T> normalized(void) const;

		// normalize() : Normalize current vector
		inline void normalize(void);

		// arithmetic operations
		inline Vector4<T> operator+( const Vector4<T>& v ) const;
		inline Vector4<T> operator-( const Vector4<T>& v ) const;
		inline Vector4<T> operator*( const Vector4<T>& v ) const;
		inline Vector4<T> operator/( const Vector4<T>& v ) const;
		inline Vector4<T> operator*( const T& s ) const;
		inline Vector4<T> operator/( const T& s ) const;
		inline Vector4<T> operator-() const;
		inline const Vector4<T>& operator=(const Vector4<T>& v);

		// operator == 
		bool operator==( const Vector4<T>& a ) const;

		// toString() : is replacsed with operator <<
		// std::string toString(void) const;


		union {
			struct {
				T x,y,z,w;
			};		
			T v[4];
		};

	};


	//////////////////////////////////////////////////////////////////////////
	// member functions

	// length()
	template<typename T>
	inline T Vector4<T>::length(void) const
	{
		return sqrt(x*x + y*y + z*z + w*w);
	}

	// squaredLength()
	template<typename T>
	inline T Vector4<T>::squaredLength(void) const
	{
		return x*x + y*y + z*z + w*w;
	}
	
	// normalized()
	template<typename T>
	inline Vector4<T> Vector4<T>::normalized(void) const
	{
		T l=length();
		return Vector4<T>(x/l,y/l,z/l,w/l);
	}

	// normalize()
	template<typename T>
	inline void Vector4<T>::normalize(void)
	{
		T l=length();
		x/=l;
		y/=l;
		z/=l;
		w/=l;
	}

	// operator +
	template<typename T>
	inline Vector4<T> Vector4<T>::operator+( const Vector4<T>& v ) const
	{
		return Vector4<T>( x+v.x, y+v.y, z+v.z , w+v.w );
	}

	// operator -
	template<typename T>
	inline Vector4<T> Vector4<T>::operator-( const Vector4<T>& v ) const
	{
		return Vector4<T>( x-v.x, y-v.y, z-v.z , w-v.w );
	}

	// operator *
	template<typename T>
	inline Vector4<T> Vector4<T>::operator*( const Vector4<T>& v ) const
	{
		return Vector4<T>( x*v.x, y*v.y, z*v.z , w*v.w );
	}

	// operator /
	template<typename T>
	inline Vector4<T> Vector4<T>::operator/( const Vector4<T>& v ) const
	{
		return Vector4<T>( x/v.x, y/v.y, z/v.z , w/v.w );
	}

	// operator * for scalars
	template<typename T>
	inline Vector4<T> Vector4<T>::operator*( const T& s ) const
	{
		return Vector4<T>( x*s, y*s, z*s, w*s );
	}

	// operator / for scalars
	template<typename T>
	inline Vector4<T> Vector4<T>::operator/( const T& s ) const
	{
		return Vector4<T>( x/s, y/s, z/s, w/s );
	}
	
	// operator - ( -v )
	template<typename T>
	inline Vector4<T> Vector4<T>::operator-() const
	{
		return Vector4<T>( -x, -y, -z, -w );
	}

	// assignment operator
	template<typename T>
	inline const Vector4<T>& Vector4<T>::operator=(const Vector4<T>& v)
	{
		x=v.x;  y=v.y;  z=v.z;  w=v.w;
		return *this;
	}

	// operator ==
	template<typename T>
	inline bool Vector4<T>::operator==( const Vector4<T>& v ) const
	{
		return (v.x==x)&&(v.y==y)&&(v.z==z)&&(v.w==w);
	}



	//////////////////////////////////////////////////////////////////////////
	// functions outside the class

	// Scalar * Vector4
	template<typename T>
	inline cvl::Vector4<T> operator*( const T& scalar, const cvl::Vector4<T> &v )
	{
		return Vector4<T> ( scalar*v.x, scalar*v.y, scalar*v.z , scalar*v.w );
	}

	// Scalar / Vector4
	template<typename T>
	inline cvl::Vector4<T> operator/( const T& scalar, const cvl::Vector4<T> &v )
	{
		return Vector4<T> ( scalar/v.x, scalar/v.y, scalar/v.z, scalar/v.w );
	}

	//////////////////////////////////////////////////////////////////////////
	// Type definition and template instantiation

#ifdef _WIN32
	// instantiation for some types 
	Vector4<float>;
	Vector4<double>; 
#endif

	typedef Vector4<float> Point4F;
	typedef Vector4<double> Point4D;

	typedef Vector4<float> Vector4F;
	typedef Vector4<double> Vector4D;

}
