﻿// Copyright © 2007-2014, Johan Hedborg, Linköping University
//                        Andreas Robinson, Linköping University
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from this
// software without specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

#include "stdafx.h"
#include "Matrix3x4.hpp"
#include "Vector2.hpp"
#include "Triangulator.h"
#include "ProjectiveAlgebra.hpp"

using namespace std;
using namespace cvl;

double g_SolverSetupTime; // Triangulation setup time
double g_SolverRuntime; // Triangulation runtime, without setup (memory allocation etc)

struct TriangulationRun
{
	// Input

	Matrix3x4D Cam0, Cam1, Cam2;
	Vector2D Point0, Point1, Point2;
	Vector3D Point3D_GT;

	// Output

	Vector3D Point3D;

	// Errors

	double Error3D;
	double ReprojErr0_2D;
	double ReprojErr1_2D;
	double ReprojErr2_2D;

	// Timing

	// (TODO)

	// Methods

	double ReprojectionError(const Matrix3x4D &cam, const Vector3D& pt3d, const Vector2D& pt2d)
	{
		return (toInhomogeneous(cam * toHomogeneous(pt3d)) - pt2d).length();
	}

	TriangulationRun()
	{
	}


	TriangulationRun(Matrix3x4D& cam0, Matrix3x4D& cam1, Matrix3x4D& cam2,
		Vector2D pt0, Vector2D pt1, Vector2D pt2, Vector3D pt3d_GT)
	{
		Cam0 = cam0;
		Cam1 = cam1;
		Cam2 = cam2;

		Point0 = pt0;
		Point1 = pt1;
		Point2 = pt2;

		Point3D_GT = pt3d_GT;
	}

	void SetResult(Vector3D pt3d)
	{
		Point3D = pt3d;
		Error3D = (Point3D - Point3D_GT).length();
		ReprojErr0_2D = ReprojectionError(Cam0, Point3D, Point0);
		ReprojErr1_2D = ReprojectionError(Cam1, Point3D, Point1);
		ReprojErr2_2D = ReprojectionError(Cam2, Point3D, Point2);

	}
};

struct TriangulationDataset
{
	// Inputs

	vector<Matrix3x4D> Cam0s;
	vector<Matrix3x4D> Cam1s;
	vector<Matrix3x4D> Cam2s;

	vector<Vector2D> Point0s;
	vector<Vector2D> Point1s;
	vector<Vector2D> Point2s;

	vector<Vector3D> Points3D_GT;

	static bool IsComma(char c)
	{
		return c == ',';
	}

	static vector<double> ReadRow(istream& in)
	{
		string line;
		vector<double> values;

		// Read line and remove commas.

		if (!getline(in, line))
			return values; // empty vector

		line.erase(std::remove_if(line.begin(), line.end(), IsComma), line.end()); // Ignore commas

		// Tokenize and convert to doubles. Note: atof() is *much* faster than iostreams on MSVC.

		stringstream sstr(line);
		for (auto sii = istream_iterator<string>(sstr); sii != istream_iterator<string>(); ++sii) {
			values.push_back(atof(sii->c_str()));
		}

		return values;
	}

	// Read a matrix from a text file
	static vector<double> LoadMatrix(const string& fname, size_t& rows, size_t& cols)
	{
		vector<double> values;
		ifstream input(fname.c_str());
		string line;

		rows = 0;
		cols = 0;

		vector<double> row_values;

		while((row_values = ReadRow(input)).size() > 0) {

			if (rows == 0) {
				cols = row_values.size();
			} else {
				if (cols != row_values.size()) {
					fprintf(stderr, "Error reading %s: expected %d columns on row %d, got %d\nExiting.\n",
						fname.c_str(), cols, rows+1, row_values.size());
					exit(0);
				}
			}

			values.insert(values.end(), row_values.begin(), row_values.end());
			rows++;
		} 

		return values;
	}

	template<int WIDTH, typename MatT>
	static vector<MatT> LoadMatrices(string fname)
	{
		vector<double> large_mat;
		vector<MatT> small_mats;
		size_t rows, cols;

		// Load large matrix

		large_mat = LoadMatrix(fname, rows, cols);
		if (cols == 0) {
			fprintf(stderr, "Failed loading file %s\nError: %s\nExiting.\n", fname.c_str(), strerror(errno));
			exit(0);
		}
		assert(cols == WIDTH);

		// Reshape the large matrix into a vector of smaller matrices
		// Each small matrix is populated with values from one row in the large matrix.

		for (size_t i = 0; i < rows; i++) {
			small_mats.push_back(MatT(&large_mat[i*cols]));
		}

		return small_mats;
	}

	void Load(string pathname, string setName)
	{
		Cam0s = LoadMatrices<12, Matrix3x4D>(pathname + "/Cam0s_" + setName + ".mat");
		Cam1s = LoadMatrices<12, Matrix3x4D>(pathname + "/Cam1s_" + setName + ".mat");
		Cam2s = LoadMatrices<12, Matrix3x4D>(pathname + "/Cam2s_" + setName + ".mat");

		Point0s = LoadMatrices<2, Vector2D>(pathname + "/Point0s_" + setName + ".mat");
		Point1s = LoadMatrices<2, Vector2D>(pathname + "/Point1s_" + setName + ".mat");
		Point2s = LoadMatrices<2, Vector2D>(pathname + "/Point2s_" + setName + ".mat");

		Points3D_GT = LoadMatrices<3, Vector3D>(pathname + "/Points3D_GT_" + setName + ".mat");
	}

	TriangulationRun GetRunData(size_t i)
	{
		return TriangulationRun(Cam0s[i], Cam1s[i], Cam2s[i], Point0s[i], Point1s[i], Point2s[i], Points3D_GT[i]);
	}
};

void SaveRuns(const vector<TriangulationRun>& runs, const string& pathname, const string& methodName, const string& setName)
{
	vector<double> Errors3D;
	vector<double> Errors2D;

	double meanError3D = 0.0;
	double varError3D = 0.0;

	size_t N = runs.size();

	for (size_t i = 0; i < N; i++) {

		double err3D = runs[i].Error3D;

		Errors3D.push_back(err3D);
		Errors2D.push_back(runs[i].ReprojErr0_2D);
		Errors2D.push_back(runs[i].ReprojErr1_2D);
		Errors2D.push_back(runs[i].ReprojErr2_2D);

		meanError3D += err3D;
		varError3D += err3D * err3D;
	}
	meanError3D /= N;
	varError3D /= N;
	varError3D = varError3D - meanError3D * meanError3D;

	fprintf(stderr, "  3D error, mu = %f, 2sigma = %f\n\n", meanError3D, 2*sqrt(varError3D));
	
	string fname = pathname + "/errors3D_" + methodName + "_" + setName + ".mat";
	ofstream e3d_file(fname.c_str());
	std::copy(Errors3D.begin(), Errors3D.end(), ostream_iterator<double>(e3d_file, ", "));
	e3d_file.close();
}

void DoDoglegTriangulation(Triangulator& solver, TriangulationRun& run)
{
	vector<Matrix3x4D> Ps;

	Ps.push_back(run.Cam0);
	Ps.push_back(run.Cam1);
	Ps.push_back(run.Cam2);

	vector<Vector2D> ps;

	ps.push_back(run.Point0);
	ps.push_back(run.Point1);
	ps.push_back(run.Point2);

	Vector3D X = solver.calc(Ps, ps);

	run.SetResult(X);
}

bool IsNonZero(int v)
{
	return v != 0;
}

int main(int argc, char* argv[])
{
	string inputPath = "./datasets";
	string outputPath = "./results";

	vector<string> setNames;

	setNames.push_back("Corridor");
	setNames.push_back("Dinosaur");
	setNames.push_back("Forward");
	setNames.push_back("ForwardNoNoise");
	//setNames.push_back("NotreDame"); // removed to reduce the size of the code archive
	setNames.push_back("Orbital");
	setNames.push_back("Sidways");

	for (auto ii = begin(setNames); ii != end(setNames); ++ii) {

		TriangulationDataset dset;
		vector<TriangulationRun> runs;

		string& setName = *ii;

		fprintf(stderr, "Working on dataset %s.\n", setName.c_str());

		fprintf(stderr, "  Loading ...\r");
		dset.Load(inputPath, setName);

		//fprintf(stderr, "  done.      \r");
		//getc(stdin);

		fprintf(stderr, "  Triangulating ...\r", setName.c_str());

		Triangulator doglegTriangulator;

		double Trun = 0;
		double Tsetup = 0;
		size_t N = dset.Point0s.size();

		for (size_t i = 0; i < N; i++) {
			TriangulationRun run = dset.GetRunData(i);
			DoDoglegTriangulation(doglegTriangulator, run);
			runs.push_back(run);
			Tsetup += g_SolverSetupTime;
			Trun += g_SolverRuntime;
		}

		fprintf(stderr, "  Triangulated %ld points in %f ms - %f us/pt\n", N, Trun * 1e3, Trun * 1e6 / N);
		fprintf(stderr, "  Setup time is %f ms - %f us/pt\n", Tsetup * 1e3, Tsetup * 1e6 / N);

		fprintf(stderr, "  Iterations required by solver:\n");
		vector<int>& H = doglegTriangulator.solver.iterHistogram;
		for (size_t i = 0; i < H.size(); i++) {
			fprintf(stderr, "      %d iterations used for %d points\n", i+1, H[i]);

			if (count_if(H.begin() + i, H.end(), IsNonZero) == 0) { // Break if remaining histogram bins are empty
				break;
			}
		}

		SaveRuns(runs, outputPath, "Ours", setName);
	}

	return 0;
}

